from __future__ import annotations

CONTENT_TYPE = ("Content-Type", "application/json; charset=utf-8")
AUTHORIZATION = ("Authorization", "Bearer %s")
X_REQUEST_ID = ("X-Request-Id", "%s")


def create_headers(
    token: str | None = None,
    with_content: bool = False,
    request_id: str | None = None,
) -> dict[str, str]:
    headers: dict[str, str] = {}

    if token:
        headers.update([(AUTHORIZATION[0], AUTHORIZATION[1] % token)])
    if with_content:
        headers.update([CONTENT_TYPE])
    if request_id:
        headers.update([(X_REQUEST_ID[0], X_REQUEST_ID[1] % request_id)])

    return headers



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_REQUEST_ID, REST_API_BASE_URL
from tests.utils.test_utils import assert_auth_header, assert_request_id_header

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import Label


@pytest.mark.asyncio
async def test_get_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
):
    label_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/labels/{label_id}"

    requests_mock.add(
        responses.GET,
        expected_endpoint,
        json=default_label_response,
        status=200,
    )

    label = todoist_api.get_label(label_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert label == default_label

    label = await todoist_api_async.get_label(label_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert label == default_label


@pytest.mark.asyncio
async def test_get_labels(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_labels_response: list[dict[str, Any]],
    default_labels_list: list[Label],
):
    requests_mock.add(
        responses.GET,
        f"{REST_API_BASE_URL}/labels",
        json=default_labels_response,
        status=200,
    )

    labels = todoist_api.get_labels()

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert labels == default_labels_list

    labels = await todoist_api_async.get_labels()

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert labels == default_labels_list


@pytest.mark.asyncio
async def test_add_label_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
):
    label_name = "A Label"
    expected_payload = {"name": label_name}

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/labels",
        json=default_label_response,
        status=200,
    )

    new_label = todoist_api.add_label(name=label_name, request_id=DEFAULT_REQUEST_ID)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_label == default_label

    new_label = await todoist_api_async.add_label(
        name=label_name, request_id=DEFAULT_REQUEST_ID
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_label == default_label


@pytest.mark.asyncio
async def test_add_label_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_label_response: dict[str, Any],
    default_label: Label,
):
    label_name = "A Label"

    optional_args = {
        "color": 30,
        "order": 3,
        "favorite": True,
    }

    expected_payload: dict[str, Any] = {"name": label_name}
    expected_payload.update(optional_args)

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/labels",
        json=default_label_response,
        status=200,
    )

    new_label = todoist_api.add_label(
        name=label_name, request_id=DEFAULT_REQUEST_ID, **optional_args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_label == default_label

    new_label = await todoist_api_async.add_label(
        name=label_name, request_id=DEFAULT_REQUEST_ID, **optional_args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_label == default_label


@pytest.mark.asyncio
async def test_update_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    label_id = "123"

    args = {
        "name": "An updated label",
        "order": 2,
        "color": 31,
        "favorite": False,
    }

    requests_mock.add(
        responses.POST, f"{REST_API_BASE_URL}/labels/{label_id}", status=204
    )

    response = todoist_api.update_label(
        label_id=label_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(args)
    assert response is True

    response = await todoist_api_async.update_label(
        label_id=label_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(args)
    assert response is True


@pytest.mark.asyncio
async def test_delete_label(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    label_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/labels/{label_id}"

    requests_mock.add(
        responses.DELETE,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.delete_label(label_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.delete_label(label_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_REQUEST_ID, REST_API_BASE_URL
from tests.utils.test_utils import assert_auth_header, assert_request_id_header

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import Section


@pytest.mark.asyncio
async def test_get_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_section_response: dict[str, Any],
    default_section: Section,
):
    section_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/sections/{section_id}"

    requests_mock.add(
        responses.GET,
        expected_endpoint,
        json=default_section_response,
        status=200,
    )

    section = todoist_api.get_section(section_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert section == default_section

    section = await todoist_api_async.get_section(section_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert section == default_section


@pytest.mark.asyncio
async def test_get_all_sections(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[dict[str, Any]],
    default_sections_list: list[Section],
):
    requests_mock.add(
        responses.GET,
        f"{REST_API_BASE_URL}/sections",
        json=default_sections_response,
        status=200,
    )

    sections = todoist_api.get_sections()

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert sections == default_sections_list

    sections = await todoist_api_async.get_sections()

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert sections == default_sections_list


@pytest.mark.asyncio
async def test_get_project_sections(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_sections_response: list[dict[str, Any]],
):
    project_id = "123"

    requests_mock.add(
        responses.GET,
        f"{REST_API_BASE_URL}/sections?project_id={project_id}",
        json=default_sections_response,
        status=200,
    )

    todoist_api.get_sections(project_id=project_id)
    await todoist_api_async.get_sections(project_id=project_id)

    assert len(requests_mock.calls) == 2


@pytest.mark.asyncio
async def test_add_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_section_response: dict[str, Any],
    default_section: Section,
):
    section_name = "A Section"
    project_id = "123"
    order = 3

    expected_payload: dict[str, Any] = {
        "name": section_name,
        "project_id": project_id,
        "order": order,
    }

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/sections",
        json=default_section_response,
        status=200,
    )

    new_section = todoist_api.add_section(
        name=section_name,
        project_id=project_id,
        order=order,
        request_id=DEFAULT_REQUEST_ID,
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_section == default_section

    new_section = await todoist_api_async.add_section(
        name=section_name,
        project_id=project_id,
        order=order,
        request_id=DEFAULT_REQUEST_ID,
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_section == default_section


@pytest.mark.asyncio
async def test_update_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    section_id = "123"

    args = {
        "name": "An updated section",
    }

    requests_mock.add(
        responses.POST, f"{REST_API_BASE_URL}/sections/{section_id}", status=204
    )

    response = todoist_api.update_section(
        section_id=section_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(args)
    assert response is True

    response = await todoist_api_async.update_section(
        section_id=section_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(args)
    assert response is True


@pytest.mark.asyncio
async def test_delete_section(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    section_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/sections/{section_id}"

    requests_mock.add(
        responses.DELETE,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.delete_section(section_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.delete_section(section_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.quick_add_responses import QUICK_ADD_RESPONSE_FULL
from tests.data.test_defaults import (
    DEFAULT_AUTH_RESPONSE,
    DEFAULT_COLLABORATORS_RESPONSE,
    DEFAULT_COMMENT_RESPONSE,
    DEFAULT_COMMENTS_RESPONSE,
    DEFAULT_COMPLETED_ITEMS_RESPONSE,
    DEFAULT_LABEL_RESPONSE,
    DEFAULT_LABELS_RESPONSE,
    DEFAULT_PROJECT_RESPONSE,
    DEFAULT_PROJECTS_RESPONSE,
    DEFAULT_SECTION_RESPONSE,
    DEFAULT_SECTIONS_RESPONSE,
    DEFAULT_TASK_RESPONSE,
    DEFAULT_TASKS_RESPONSE,
    DEFAULT_TOKEN,
)
from todoist_api_python.api import TodoistAPI
from todoist_api_python.api_async import TodoistAPIAsync
from todoist_api_python.models import (
    AuthResult,
    Collaborator,
    Comment,
    CompletedItems,
    Label,
    Project,
    QuickAddResult,
    Section,
    Task,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def requests_mock() -> Iterator[responses.RequestsMock]:
    with responses.RequestsMock() as requests_mock:
        yield requests_mock


@pytest.fixture
def todoist_api() -> TodoistAPI:
    return TodoistAPI(DEFAULT_TOKEN)


@pytest.fixture
def todoist_api_async() -> TodoistAPIAsync:
    return TodoistAPIAsync(DEFAULT_TOKEN)


@pytest.fixture
def default_task_response() -> dict[str, Any]:
    return DEFAULT_TASK_RESPONSE


@pytest.fixture
def default_task() -> Task:
    return Task.from_dict(DEFAULT_TASK_RESPONSE)


@pytest.fixture
def default_tasks_response() -> list[dict[str, Any]]:
    return DEFAULT_TASKS_RESPONSE


@pytest.fixture
def default_tasks_list() -> list[Task]:
    return [Task.from_dict(obj) for obj in DEFAULT_TASKS_RESPONSE]


@pytest.fixture
def default_project_response() -> dict[str, Any]:
    return DEFAULT_PROJECT_RESPONSE


@pytest.fixture
def default_project() -> Project:
    return Project.from_dict(DEFAULT_PROJECT_RESPONSE)


@pytest.fixture
def default_projects_response() -> list[dict[str, Any]]:
    return DEFAULT_PROJECTS_RESPONSE


@pytest.fixture
def default_projects_list() -> list[Project]:
    return [Project.from_dict(obj) for obj in DEFAULT_PROJECTS_RESPONSE]


@pytest.fixture
def default_collaborators_response() -> list[dict[str, Any]]:
    return DEFAULT_COLLABORATORS_RESPONSE


@pytest.fixture
def default_collaborators_list() -> list[Collaborator]:
    return [Collaborator.from_dict(obj) for obj in DEFAULT_COLLABORATORS_RESPONSE]


@pytest.fixture
def default_section_response() -> dict[str, Any]:
    return DEFAULT_SECTION_RESPONSE


@pytest.fixture
def default_section() -> Section:
    return Section.from_dict(DEFAULT_SECTION_RESPONSE)


@pytest.fixture
def default_sections_response() -> list[dict[str, Any]]:
    return DEFAULT_SECTIONS_RESPONSE


@pytest.fixture
def default_sections_list() -> list[Section]:
    return [Section.from_dict(obj) for obj in DEFAULT_SECTIONS_RESPONSE]


@pytest.fixture
def default_comment_response() -> dict[str, Any]:
    return DEFAULT_COMMENT_RESPONSE


@pytest.fixture
def default_comment() -> Comment:
    return Comment.from_dict(DEFAULT_COMMENT_RESPONSE)


@pytest.fixture
def default_comments_response() -> list[dict[str, Any]]:
    return DEFAULT_COMMENTS_RESPONSE


@pytest.fixture
def default_comments_list() -> list[Comment]:
    return [Comment.from_dict(obj) for obj in DEFAULT_COMMENTS_RESPONSE]


@pytest.fixture
def default_label_response() -> dict[str, Any]:
    return DEFAULT_LABEL_RESPONSE


@pytest.fixture
def default_label() -> Label:
    return Label.from_dict(DEFAULT_LABEL_RESPONSE)


@pytest.fixture
def default_labels_response() -> list[dict[str, Any]]:
    return DEFAULT_LABELS_RESPONSE


@pytest.fixture
def default_labels_list() -> list[Label]:
    return [Label.from_dict(obj) for obj in DEFAULT_LABELS_RESPONSE]


@pytest.fixture
def default_quick_add_response() -> dict[str, Any]:
    return QUICK_ADD_RESPONSE_FULL


@pytest.fixture
def default_quick_add_result() -> QuickAddResult:
    return QuickAddResult.from_quick_add_response(QUICK_ADD_RESPONSE_FULL)


@pytest.fixture
def default_auth_response() -> dict[str, Any]:
    return DEFAULT_AUTH_RESPONSE


@pytest.fixture
def default_auth_result() -> AuthResult:
    return AuthResult.from_dict(DEFAULT_AUTH_RESPONSE)


@pytest.fixture
def default_completed_items_response() -> dict[str, Any]:
    return DEFAULT_COMPLETED_ITEMS_RESPONSE


@pytest.fixture
def default_completed_items() -> CompletedItems:
    return CompletedItems.from_dict(DEFAULT_COMPLETED_ITEMS_RESPONSE)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import pytest
import responses

from tests.data.test_defaults import DEFAULT_REQUEST_ID, REST_API_BASE_URL
from tests.utils.test_utils import assert_auth_header, assert_request_id_header

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import Comment


@pytest.mark.asyncio
async def test_get_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comment_response: dict[str, Any],
    default_comment: Comment,
):
    comment_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/comments/{comment_id}"

    requests_mock.add(
        responses.GET,
        expected_endpoint,
        json=default_comment_response,
        status=200,
    )

    comment = todoist_api.get_comment(comment_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert comment == default_comment

    comment = await todoist_api_async.get_comment(comment_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert comment == default_comment


@pytest.mark.asyncio
async def test_get_comments(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comments_response: list[dict[str, Any]],
    default_comments_list: list[Comment],
):
    task_id = "1234"

    requests_mock.add(
        responses.GET,
        f"{REST_API_BASE_URL}/comments?task_id={task_id}",
        json=default_comments_response,
        status=200,
    )

    comments = todoist_api.get_comments(task_id=task_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert comments == default_comments_list

    comments = await todoist_api_async.get_comments(task_id=task_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert comments == default_comments_list


@pytest.mark.asyncio
async def test_add_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_comment_response: dict[str, Any],
    default_comment: Comment,
):
    content = "A Comment"
    project_id = 123
    attachment_data = {
        "resource_type": "file",
        "file_url": "https://s3.amazonaws.com/domorebetter/Todoist+Setup+Guide.pdf",
        "file_type": "application/pdf",
        "file_name": "File.pdf",
    }

    expected_payload: dict[str, Any] = {
        "content": content,
        "project_id": project_id,
        "attachment": attachment_data,
    }

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/comments",
        json=default_comment_response,
        status=200,
    )

    new_comment = todoist_api.add_comment(
        content=content,
        project_id=project_id,
        attachment=attachment_data,
        request_id=DEFAULT_REQUEST_ID,
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_comment == default_comment

    new_comment = await todoist_api_async.add_comment(
        content=content,
        project_id=project_id,
        attachment=attachment_data,
        request_id=DEFAULT_REQUEST_ID,
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_comment == default_comment


@pytest.mark.asyncio
async def test_update_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    comment_id = "1234"

    args = {
        "content": "An updated comment",
    }

    requests_mock.add(
        responses.POST, f"{REST_API_BASE_URL}/comments/{comment_id}", status=204
    )

    response = todoist_api.update_comment(
        comment_id=comment_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(args)
    assert response is True

    response = await todoist_api_async.update_comment(
        comment_id=comment_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(args)
    assert response is True


@pytest.mark.asyncio
async def test_delete_comment(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    comment_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/comments/{comment_id}"

    requests_mock.add(
        responses.DELETE,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.delete_comment(comment_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.delete_comment(comment_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

import pytest
import responses

from tests.data.test_defaults import (
    DEFAULT_REQUEST_ID,
    REST_API_BASE_URL,
    SYNC_API_BASE_URL,
)
from tests.utils.test_utils import assert_auth_header, assert_request_id_header

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import QuickAddResult, Task


@pytest.mark.asyncio
async def test_get_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
):
    task_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/tasks/{task_id}"

    requests_mock.add(
        responses.GET,
        expected_endpoint,
        json=default_task_response,
        status=200,
    )

    task = todoist_api.get_task(task_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert task == default_task

    task = await todoist_api_async.get_task(task_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert task == default_task


@pytest.mark.asyncio
async def test_get_tasks_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_tasks_response: list[dict[str, Any]],
    default_tasks_list: list[Task],
):
    requests_mock.add(
        responses.GET,
        f"{REST_API_BASE_URL}/tasks",
        json=default_tasks_response,
        status=200,
    )

    tasks = todoist_api.get_tasks()

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert tasks == default_tasks_list

    tasks = await todoist_api_async.get_tasks()

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert tasks == default_tasks_list


@pytest.mark.asyncio
async def test_get_tasks_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_tasks_response: list[dict[str, Any]],
    default_tasks_list: list[Task],
):
    project_id = "1234"
    label_id = 2345
    filter = "today"
    lang = "en"
    ids = [1, 2, 3, 4]

    encoded_ids = quote(",".join(str(x) for x in ids))
    expected_endpoint = (
        f"{REST_API_BASE_URL}/tasks"
        f"?project_id={project_id}&label_id={label_id}"
        f"&filter={filter}&lang={lang}&ids={encoded_ids}"
    )

    requests_mock.add(
        responses.GET, expected_endpoint, json=default_tasks_response, status=200
    )

    tasks = todoist_api.get_tasks(
        project_id=project_id, label_id=label_id, filter=filter, lang=lang, ids=ids
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert tasks == default_tasks_list

    tasks = await todoist_api_async.get_tasks(
        project_id=project_id, label_id=label_id, filter=filter, lang=lang, ids=ids
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert tasks == default_tasks_list


@pytest.mark.asyncio
async def test_add_task_minimal(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
):
    task_content = "Some content"
    expected_payload = {"content": task_content}

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/tasks",
        json=default_task_response,
        status=200,
    )

    new_task = todoist_api.add_task(content=task_content, request_id=DEFAULT_REQUEST_ID)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_task == default_task

    new_task = await todoist_api_async.add_task(
        content=task_content, request_id=DEFAULT_REQUEST_ID
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_task == default_task


@pytest.mark.asyncio
async def test_add_task_full(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_task_response: dict[str, Any],
    default_task: Task,
):
    task_content = "Some content"

    optional_args = {
        "description": "A description",
        "project_id": 123,
        "section_id": 456,
        "parent_id": 789,
        "order": 3,
        "label_ids": [123, 456],
        "priority": 4,
        "due_string": "today",
        "due_date": "2021-01-01",
        "due_datetime": "2021-01-01T11:00:00Z",
        "due_lang": "ja",
        "assignee": 321,
    }

    expected_payload: dict[str, Any] = {"content": task_content}
    expected_payload.update(optional_args)

    requests_mock.add(
        responses.POST,
        f"{REST_API_BASE_URL}/tasks",
        json=default_task_response,
        status=200,
    )

    new_task = todoist_api.add_task(
        content=task_content, request_id=DEFAULT_REQUEST_ID, **optional_args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert new_task == default_task

    new_task = await todoist_api_async.add_task(
        content=task_content, request_id=DEFAULT_REQUEST_ID, **optional_args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert new_task == default_task


@pytest.mark.asyncio
async def test_update_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    task_id = "123"

    args = {
        "content": "Some updated content",
        "description": "An updated description",
        "label_ids": ["123", "456"],
        "priority": 4,
        "due_string": "today",
        "due_date": "2021-01-01",
        "due_datetime": "2021-01-01T11:00:00Z",
        "due_lang": "ja",
        "assignee": "321",
    }

    requests_mock.add(
        responses.POST, f"{REST_API_BASE_URL}/tasks/{task_id}", status=204
    )

    response = todoist_api.update_task(
        task_id=task_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_request_id_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(args)
    assert response is True

    response = await todoist_api_async.update_task(
        task_id=task_id, request_id=DEFAULT_REQUEST_ID, **args
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_request_id_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(args)
    assert response is True


@pytest.mark.asyncio
async def test_close_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    task_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/tasks/{task_id}/close"

    requests_mock.add(
        responses.POST,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.close_task(task_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.close_task(task_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True


@pytest.mark.asyncio
async def test_reopen_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    task_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/tasks/{task_id}/reopen"

    requests_mock.add(
        responses.POST,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.reopen_task(task_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.reopen_task(task_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True


@pytest.mark.asyncio
async def test_delete_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
):
    task_id = "1234"
    expected_endpoint = f"{REST_API_BASE_URL}/tasks/{task_id}"

    requests_mock.add(
        responses.DELETE,
        expected_endpoint,
        status=204,
    )

    response = todoist_api.delete_task(task_id)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert response is True

    response = await todoist_api_async.delete_task(task_id)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert response is True


@pytest.mark.asyncio
async def test_quick_add_task(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_quick_add_response: dict[str, Any],
    default_quick_add_result: QuickAddResult,
):
    text = "some task"
    expected_payload = {"text": text, "meta": True, "auto_reminder": True}

    requests_mock.add(
        responses.POST,
        f"{SYNC_API_BASE_URL}/quick/add",
        json=default_quick_add_response,
        status=200,
    )

    response = todoist_api.quick_add_task(text=text)

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert requests_mock.calls[0].request.body == json.dumps(expected_payload)
    assert response == default_quick_add_result

    response = await todoist_api_async.quick_add_task(text=text)

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert requests_mock.calls[1].request.body == json.dumps(expected_payload)
    assert response == default_quick_add_result



from __future__ import annotations

import json
from typing import Any

import pytest
import responses
from requests import HTTPError, Session

from tests.conftest import DEFAULT_TOKEN
from todoist_api_python.endpoints import BASE_URL, TASKS_ENDPOINT
from todoist_api_python.http_requests import delete, get, post

DEFAULT_URL = f"{BASE_URL}/{TASKS_ENDPOINT}"


@responses.activate
def test_get_with_params(default_task_response: dict[str, Any]):
    params = {"param1": "value1", "param2": "value2"}

    responses.add(
        responses.GET,
        DEFAULT_URL,
        json=default_task_response,
        status=200,
    )

    response = get(Session(), DEFAULT_URL, DEFAULT_TOKEN, params)

    assert len(responses.calls) == 1
    assert (
        responses.calls[0].request.url == f"{DEFAULT_URL}?param1=value1&param2=value2"
    )
    assert (
        responses.calls[0].request.headers["Authorization"] == f"Bearer {DEFAULT_TOKEN}"
    )
    assert response == default_task_response


@responses.activate
def test_get_raise_for_status():
    with pytest.raises(HTTPError):
        responses.add(
            responses.GET,
            DEFAULT_URL,
            status=500,
        )

        get(Session(), DEFAULT_URL, DEFAULT_TOKEN)


@responses.activate
def test_post_with_data(default_task_response: dict[str, Any]):
    request_id = "12345"

    data = {"param1": "value1", "param2": "value2", "request_id": request_id}

    responses.add(
        responses.POST,
        DEFAULT_URL,
        json=default_task_response,
        status=200,
    )

    response = post(Session(), DEFAULT_URL, DEFAULT_TOKEN, data)

    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == DEFAULT_URL
    assert (
        responses.calls[0].request.headers["Authorization"] == f"Bearer {DEFAULT_TOKEN}"
    )
    assert responses.calls[0].request.headers["X-Request-Id"] == request_id
    assert (
        responses.calls[0].request.headers["Content-Type"]
        == "application/json; charset=utf-8"
    )
    assert responses.calls[0].request.body == json.dumps(data)
    assert response == default_task_response


@responses.activate
def test_post_return_ok_when_no_response_body():
    responses.add(
        responses.POST,
        DEFAULT_URL,
        status=204,
    )

    result = post(Session(), DEFAULT_URL, DEFAULT_TOKEN)

    assert result is True


@responses.activate
def test_post_raise_for_status():
    with pytest.raises(HTTPError):
        responses.add(
            responses.POST,
            DEFAULT_URL,
            status=500,
        )

        post(Session(), DEFAULT_URL, DEFAULT_TOKEN)


@responses.activate
def test_delete_with_request_id():
    request_id = "12345"

    responses.add(
        responses.DELETE,
        DEFAULT_URL,
        status=204,
    )

    result = delete(Session(), DEFAULT_URL, DEFAULT_TOKEN, {"request_id": request_id})

    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == DEFAULT_URL
    assert (
        responses.calls[0].request.headers["Authorization"] == f"Bearer {DEFAULT_TOKEN}"
    )
    assert responses.calls[0].request.headers["X-Request-Id"] == request_id
    assert result is True


@responses.activate
def test_delete_raise_for_status():
    with pytest.raises(HTTPError):
        responses.add(
            responses.DELETE,
            DEFAULT_URL,
            status=500,
        )

        delete(Session(), DEFAULT_URL, DEFAULT_TOKEN)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from tests.data.test_defaults import DEFAULT_REQUEST_ID, DEFAULT_TOKEN
from todoist_api_python.api import TodoistAPI

if TYPE_CHECKING:
    from collections.abc import Callable

MATCH_ANY_REGEX = re.compile(".*")


def assert_auth_header(request):
    assert request.headers["Authorization"] == f"Bearer {DEFAULT_TOKEN}"


def assert_request_id_header(request):
    assert request.headers["X-Request-Id"] == DEFAULT_REQUEST_ID


def get_todoist_api_patch(method: Callable | None) -> str:
    module = TodoistAPI.__module__
    name = TodoistAPI.__qualname__

    return f"{module}.{name}.{method.__name__}" if method else f"{module}.{name}"



from __future__ import annotations

from typing import TYPE_CHECKING

from todoist_api_python.api import TodoistAPI
from todoist_api_python.utils import run_async

if TYPE_CHECKING:
    import requests

    from todoist_api_python.models import (
        Collaborator,
        Comment,
        CompletedItems,
        Label,
        Project,
        QuickAddResult,
        Section,
        Task,
    )


class TodoistAPIAsync:
    def __init__(self, token: str, session: requests.Session | None = None) -> None:
        self._api = TodoistAPI(token, session)

    async def get_task(self, task_id: str) -> Task:
        return await run_async(lambda: self._api.get_task(task_id))

    async def get_tasks(self, **kwargs) -> list[Task]:
        return await run_async(lambda: self._api.get_tasks(**kwargs))

    async def add_task(self, content: str, **kwargs) -> Task:
        return await run_async(lambda: self._api.add_task(content, **kwargs))

    async def update_task(self, task_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.update_task(task_id, **kwargs))

    async def close_task(self, task_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.close_task(task_id, **kwargs))

    async def reopen_task(self, task_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.reopen_task(task_id, **kwargs))

    async def delete_task(self, task_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.delete_task(task_id, **kwargs))

    async def quick_add_task(self, text: str) -> QuickAddResult:
        return await run_async(lambda: self._api.quick_add_task(text))

    async def get_project(self, project_id: str) -> Project:
        return await run_async(lambda: self._api.get_project(project_id))

    async def get_projects(self) -> list[Project]:
        return await run_async(lambda: self._api.get_projects())

    async def add_project(self, name: str, **kwargs) -> Project:
        return await run_async(lambda: self._api.add_project(name, **kwargs))

    async def update_project(self, project_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.update_project(project_id, **kwargs))

    async def delete_project(self, project_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.delete_project(project_id, **kwargs))

    async def get_collaborators(self, project_id: str) -> list[Collaborator]:
        return await run_async(lambda: self._api.get_collaborators(project_id))

    async def get_section(self, section_id: str) -> Section:
        return await run_async(lambda: self._api.get_section(section_id))

    async def get_sections(self, **kwargs) -> list[Section]:
        return await run_async(lambda: self._api.get_sections(**kwargs))

    async def add_section(self, name: str, project_id: str, **kwargs) -> Section:
        return await run_async(
            lambda: self._api.add_section(name, project_id, **kwargs)
        )

    async def update_section(self, section_id: str, name: str, **kwargs) -> bool:
        return await run_async(
            lambda: self._api.update_section(section_id, name, **kwargs)
        )

    async def delete_section(self, section_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.delete_section(section_id, **kwargs))

    async def get_comment(self, comment_id: str) -> Comment:
        return await run_async(lambda: self._api.get_comment(comment_id))

    async def get_comments(self, **kwargs) -> list[Comment]:
        return await run_async(lambda: self._api.get_comments(**kwargs))

    async def add_comment(self, content: str, **kwargs) -> Comment:
        return await run_async(lambda: self._api.add_comment(content, **kwargs))

    async def update_comment(self, comment_id: str, content: str, **kwargs) -> bool:
        return await run_async(
            lambda: self._api.update_comment(comment_id, content, **kwargs)
        )

    async def delete_comment(self, comment_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.delete_comment(comment_id, **kwargs))

    async def get_label(self, label_id: str) -> Label:
        return await run_async(lambda: self._api.get_label(label_id))

    async def get_labels(self) -> list[Label]:
        return await run_async(lambda: self._api.get_labels())

    async def add_label(self, name: str, **kwargs) -> Label:
        return await run_async(lambda: self._api.add_label(name, **kwargs))

    async def update_label(self, label_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.update_label(label_id, **kwargs))

    async def delete_label(self, label_id: str, **kwargs) -> bool:
        return await run_async(lambda: self._api.delete_label(label_id, **kwargs))

    async def get_shared_labels(self) -> list[str]:
        return await run_async(lambda: self._api.get_shared_labels())

    async def rename_shared_label(self, name: str, new_name: str) -> bool:
        return await run_async(lambda: self._api.rename_shared_label(name, new_name))

    async def remove_shared_label(self, name: str) -> bool:
        return await run_async(lambda: self._api.remove_shared_label(name))

    async def get_completed_items(
        self,
        project_id: str | None = None,
        section_id: str | None = None,
        item_id: str | None = None,
        last_seen_id: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> CompletedItems:
        return await run_async(
            lambda: self._api.get_completed_items(
                project_id, section_id, item_id, last_seen_id, limit, cursor
            )
        )



from __future__ import annotations

import asyncio

SHOW_TASK_ENDPOINT = "https://todoist.com/showTask"


def get_url_for_task(task_id: int, sync_id: int | None) -> str:
    return (
        f"{SHOW_TASK_ENDPOINT}?id={task_id}&sync_id={sync_id}"
        if sync_id
        else f"{SHOW_TASK_ENDPOINT}?id={task_id}"
    )


async def run_async(func):
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, func)



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Any, Literal

from todoist_api_python.utils import get_url_for_task

VIEW_STYLE = Literal["list", "board"]


@dataclass
class Project:
    color: str
    comment_count: int
    id: str
    is_favorite: bool
    is_inbox_project: bool | None
    is_shared: bool
    is_team_inbox: bool | None
    can_assign_tasks: bool | None
    name: str
    order: int
    parent_id: str | None
    url: str
    view_style: VIEW_STYLE

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            color=obj["color"],
            comment_count=obj["comment_count"],
            id=obj["id"],
            is_favorite=obj["is_favorite"],
            is_inbox_project=obj.get("is_inbox_project"),
            is_shared=obj["is_shared"],
            is_team_inbox=obj.get("is_team_inbox"),
            can_assign_tasks=obj.get("can_assign_tasks"),
            name=obj["name"],
            order=obj["order"],
            parent_id=obj.get("parent_id"),
            url=obj["url"],
            view_style=obj["view_style"],
        )


@dataclass
class Section:
    id: str
    name: str
    order: int
    project_id: str

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            id=obj["id"],
            name=obj["name"],
            order=obj["order"],
            project_id=obj["project_id"],
        )


@dataclass
class Due:
    date: str
    is_recurring: bool
    string: str

    datetime: str | None = None
    timezone: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            date=obj["date"],
            is_recurring=obj["is_recurring"],
            string=obj["string"],
            datetime=obj.get("datetime"),
            timezone=obj.get("timezone"),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "date": self.date,
            "is_recurring": self.is_recurring,
            "string": self.string,
            "datetime": self.datetime,
            "timezone": self.timezone,
        }

    @classmethod
    def from_quick_add_response(cls, obj: dict[str, Any]):
        due = obj.get("due")

        if not due:
            return None

        timezone = due.get("timezone")

        datetime: str | None = None

        if timezone:
            datetime = due["date"]

        return cls(
            date=due["date"],
            is_recurring=due["is_recurring"],
            string=due["string"],
            datetime=datetime,
            timezone=timezone,
        )


@dataclass
class Task:
    assignee_id: str | None
    assigner_id: str | None
    comment_count: int
    is_completed: bool
    content: str
    created_at: str
    creator_id: str
    description: str
    due: Due | None
    id: str
    labels: list[str] | None
    order: int
    parent_id: str | None
    priority: int
    project_id: str
    section_id: str | None
    url: str
    duration: Duration | None

    sync_id: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        due: Due | None = None
        duration: Duration | None = None

        if obj.get("due"):
            due = Due.from_dict(obj["due"])

        if obj.get("duration"):
            duration = Duration.from_dict(obj["duration"])

        return cls(
            assignee_id=obj.get("assignee_id"),
            assigner_id=obj.get("assigner_id"),
            comment_count=obj["comment_count"],
            is_completed=obj["is_completed"],
            content=obj["content"],
            created_at=obj["created_at"],
            creator_id=obj["creator_id"],
            description=obj["description"],
            due=due,
            id=obj["id"],
            labels=obj.get("labels"),
            order=obj["order"],
            parent_id=obj.get("parent_id"),
            priority=obj["priority"],
            project_id=obj["project_id"],
            section_id=obj.get("section_id"),
            url=obj["url"],
            duration=duration,
        )

    def to_dict(self) -> dict[str, Any]:
        due: dict[str, Any] | None = None
        duration: dict[str, Any] | None = None

        if self.due:
            due = self.due.to_dict()

        if self.duration:
            duration = self.duration.to_dict()

        return {
            "assignee_id": self.assignee_id,
            "assigner_id": self.assigner_id,
            "comment_count": self.comment_count,
            "is_completed": self.is_completed,
            "content": self.content,
            "created_at": self.created_at,
            "creator_id": self.creator_id,
            "description": self.description,
            "due": due,
            "id": self.id,
            "labels": self.labels,
            "order": self.order,
            "parent_id": self.parent_id,
            "priority": self.priority,
            "project_id": self.project_id,
            "section_id": self.section_id,
            "sync_id": self.sync_id,
            "url": self.url,
            "duration": duration,
        }

    @classmethod
    def from_quick_add_response(cls, obj: dict[str, Any]):
        due: Due | None = None
        duration: Duration | None = None

        if obj.get("due"):
            due = Due.from_quick_add_response(obj)

        if obj.get("duration"):
            duration = Duration.from_dict(obj["duration"])

        return cls(
            assignee_id=obj.get("responsible_uid"),
            assigner_id=obj.get("assigned_by_uid"),
            comment_count=0,
            is_completed=False,
            content=obj["content"],
            created_at=obj["added_at"],
            creator_id=obj["added_by_uid"],
            description=obj["description"],
            due=due,
            duration=duration,
            id=obj["id"],
            labels=obj["labels"],
            order=obj["child_order"],
            parent_id=obj["parent_id"] or None,
            priority=obj["priority"],
            project_id=obj["project_id"],
            section_id=obj["section_id"] or None,
            sync_id=obj["sync_id"],
            url=get_url_for_task(obj["id"], obj["sync_id"]),
        )


@dataclass
class QuickAddResult:
    task: Task

    resolved_project_name: str | None = None
    resolved_assignee_name: str | None = None
    resolved_label_names: list[str] | None = None
    resolved_section_name: str | None = None

    @classmethod
    def from_quick_add_response(cls, obj: dict[str, Any]):
        project_data = obj["meta"].get("project", {})
        assignee_data = obj["meta"].get("assignee", {})
        section_data = obj["meta"].get("section", {})

        resolved_project_name = None
        resolved_assignee_name = None
        resolved_section_name = None

        if project_data and len(project_data) == 2:
            resolved_project_name = obj["meta"]["project"][1]

        if assignee_data and len(assignee_data) == 2:
            resolved_assignee_name = obj["meta"]["assignee"][1]

        if section_data and len(section_data) == 2:
            resolved_section_name = obj["meta"]["section"][1]

        return cls(
            task=Task.from_quick_add_response(obj),
            resolved_project_name=resolved_project_name,
            resolved_assignee_name=resolved_assignee_name,
            resolved_label_names=list(obj["meta"]["labels"].values()),
            resolved_section_name=resolved_section_name,
        )


@dataclass
class Collaborator:
    id: str
    email: str
    name: str

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            id=obj["id"],
            email=obj["email"],
            name=obj["name"],
        )


@dataclass
class Attachment:
    resource_type: str | None = None

    file_name: str | None = None
    file_size: int | None = None
    file_type: str | None = None
    file_url: str | None = None
    file_duration: int | None = None
    upload_state: str | None = None

    image: str | None = None
    image_width: int | None = None
    image_height: int | None = None

    url: str | None = None
    title: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            resource_type=obj.get("resource_type"),
            file_name=obj.get("file_name"),
            file_size=obj.get("file_size"),
            file_type=obj.get("file_type"),
            file_url=obj.get("file_url"),
            upload_state=obj.get("upload_state"),
            image=obj.get("image"),
            image_width=obj.get("image_width"),
            image_height=obj.get("image_height"),
            url=obj.get("url"),
            title=obj.get("title"),
        )


@dataclass
class Comment:
    attachment: Attachment | None
    content: str
    id: str
    posted_at: str
    project_id: str | None
    task_id: str | None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        attachment: Attachment | None = None

        if "attachment" in obj and obj["attachment"] is not None:
            attachment = Attachment.from_dict(obj["attachment"])

        return cls(
            attachment=attachment,
            content=obj["content"],
            id=obj["id"],
            posted_at=obj["posted_at"],
            project_id=obj.get("project_id"),
            task_id=obj.get("task_id"),
        )


@dataclass
class Label:
    id: str
    name: str
    color: str
    order: int
    is_favorite: bool

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            id=obj["id"],
            name=obj["name"],
            color=obj["color"],
            order=obj["order"],
            is_favorite=obj["is_favorite"],
        )


@dataclass
class AuthResult:
    access_token: str
    state: str | None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            access_token=obj["access_token"],
            state=obj.get("state"),
        )


@dataclass
class Item:
    id: str
    user_id: str
    project_id: str
    content: str
    description: str
    priority: int
    child_order: int
    collapsed: bool
    labels: list[str]
    checked: bool
    is_deleted: bool
    added_at: str
    due: Due | None = None
    parent_id: int | None = None
    section_id: str | None = None
    day_order: int | None = None
    added_by_uid: str | None = None
    assigned_by_uid: str | None = None
    responsible_uid: str | None = None
    sync_id: str | None = None
    completed_at: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]) -> Item:
        params = {f.name: obj[f.name] for f in fields(cls) if f.name in obj}
        if (due := obj.get("due")) is not None:
            params["due"] = Due.from_dict(due)

        return cls(**params)


@dataclass
class ItemCompletedInfo:
    item_id: str
    completed_items: int

    @classmethod
    def from_dict(cls, obj: dict[str, Any]) -> ItemCompletedInfo:
        return cls(**{f.name: obj[f.name] for f in fields(cls)})


@dataclass
class CompletedItems:
    items: list[Item]
    total: int
    completed_info: list[ItemCompletedInfo]
    has_more: bool
    next_cursor: str | None = None

    @classmethod
    def from_dict(cls, obj: dict[str, Any]) -> CompletedItems:
        return cls(
            items=[Item.from_dict(v) for v in obj["items"]],
            total=obj["total"],
            completed_info=[
                ItemCompletedInfo.from_dict(v) for v in obj["completed_info"]
            ],
            has_more=obj["has_more"],
            next_cursor=obj.get("next_cursor"),
        )


@dataclass
class Duration:
    amount: int
    unit: str

    @classmethod
    def from_dict(cls, obj: dict[str, Any]):
        return cls(
            amount=obj["amount"],
            unit=obj["unit"],
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "amount": self.amount,
            "unit": self.unit,
        }



from __future__ import annotations

from typing import Any

QUICK_ADD_RESPONSE_MINIMAL: dict[str, Any] = {
    "added_by_uid": "21180723",
    "assigned_by_uid": None,
    "checked": 0,
    "child_order": 6,
    "collapsed": 0,
    "content": "some task",
    "description": "",
    "added_at": "2021-02-05T11:02:56.00000Z",
    "date_completed": None,
    "due": None,
    "id": "4554989047",
    "in_history": 0,
    "is_deleted": 0,
    "labels": [],
    "meta": {
        "assignee": [None, None],
        "due": {
            "date_local": None,
            "datetime_local": None,
            "datetime_utc": None,
            "is_recurring": False,
            "lang": None,
            "object_type": "null",
            "string": None,
            "timezone": None,
            "timezone_name": None,
        },
        "labels": {},
        "priority": 1,
        "project": [None, None],
        "section": [None, None],
        "text": "some task",
    },
    "parent_id": None,
    "priority": 1,
    "project_id": "2203108698",
    "responsible_uid": None,
    "section_id": None,
    "sync_id": None,
    "user_id": "21180723",
}

QUICK_ADD_RESPONSE_FULL: dict[str, Any] = {
    "added_by_uid": "21180723",
    "assigned_by_uid": "21180723",
    "checked": 0,
    "child_order": 1,
    "collapsed": 0,
    "content": "some task",
    "description": "a description",
    "added_at": "2021-02-05T11:04:54.00000Z",
    "date_completed": None,
    "due": {
        "date": "2021-02-06T11:00:00.00000Z",
        "is_recurring": False,
        "lang": "en",
        "string": "Feb 6 11:00 AM",
        "timezone": "Europe/London",
    },
    "id": "4554993687",
    "in_history": 0,
    "is_deleted": 0,
    "labels": ["Label1", "Label2"],
    "meta": {
        "assignee": ["29172386", "Some Guy"],
        "due": {
            "date_local": "2021-02-06T00:00:00.00000Z",
            "datetime_local": "2021-02-06T11:00:00.00000Z",
            "datetime_utc": "2021-02-06T11:00:00.00000Z",
            "is_recurring": False,
            "lang": "en",
            "object_type": "fixed_datetime",
            "string": "Feb 6 11:00 AM",
            "timezone": {"zone": "Europe/London"},
            "timezone_name": "Europe/London",
        },
        "labels": {"2156154810": "Label1", "2156154812": "Label2"},
        "priority": 1,
        "project": ["2257514220", "test"],
        "section": ["2232454220", "A section"],
        "text": "some task",
    },
    "parent_id": None,
    "priority": 1,
    "project_id": "2257514220",
    "responsible_uid": "29172386",
    "section_id": "2232454220",
    "sync_id": "4554993687",
    "user_id": "21180723",
}



from __future__ import annotations

from typing import Any
from weakref import finalize

import requests

from todoist_api_python.endpoints import (
    COLLABORATORS_ENDPOINT,
    COMMENTS_ENDPOINT,
    COMPLETED_ITEMS_ENDPOINT,
    LABELS_ENDPOINT,
    PROJECTS_ENDPOINT,
    QUICK_ADD_ENDPOINT,
    SECTIONS_ENDPOINT,
    SHARED_LABELS_ENDPOINT,
    SHARED_LABELS_REMOVE_ENDPOINT,
    SHARED_LABELS_RENAME_ENDPOINT,
    TASKS_ENDPOINT,
    get_rest_url,
    get_sync_url,
)
from todoist_api_python.http_requests import delete, get, post
from todoist_api_python.models import (
    Collaborator,
    Comment,
    CompletedItems,
    Label,
    Project,
    QuickAddResult,
    Section,
    Task,
)


class TodoistAPI:
    def __init__(self, token: str, session: requests.Session | None = None) -> None:
        self._token: str = token
        self._session = session or requests.Session()
        self._finalizer = finalize(self, self._session.close)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._finalizer()

    def get_task(self, task_id: str) -> Task:
        endpoint = get_rest_url(f"{TASKS_ENDPOINT}/{task_id}")
        task = get(self._session, endpoint, self._token)
        return Task.from_dict(task)

    def get_tasks(self, **kwargs) -> list[Task]:
        ids = kwargs.pop("ids", None)

        if ids:
            kwargs.update({"ids": ",".join(str(i) for i in ids)})

        endpoint = get_rest_url(TASKS_ENDPOINT)
        tasks = get(self._session, endpoint, self._token, kwargs)
        return [Task.from_dict(obj) for obj in tasks]

    def add_task(self, content: str, **kwargs) -> Task:
        endpoint = get_rest_url(TASKS_ENDPOINT)
        data: dict[str, Any] = {"content": content}
        data.update(kwargs)
        task = post(self._session, endpoint, self._token, data=data)
        return Task.from_dict(task)

    def update_task(self, task_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{TASKS_ENDPOINT}/{task_id}")
        return post(self._session, endpoint, self._token, data=kwargs)

    def close_task(self, task_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{TASKS_ENDPOINT}/{task_id}/close")
        return post(self._session, endpoint, self._token, data=kwargs)

    def reopen_task(self, task_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{TASKS_ENDPOINT}/{task_id}/reopen")
        return post(self._session, endpoint, self._token, data=kwargs)

    def delete_task(self, task_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{TASKS_ENDPOINT}/{task_id}")
        return delete(self._session, endpoint, self._token, args=kwargs)

    def quick_add_task(self, text: str) -> QuickAddResult:
        endpoint = get_sync_url(QUICK_ADD_ENDPOINT)
        data = {
            "text": text,
            "meta": True,
            "auto_reminder": True,
        }
        task_data = post(self._session, endpoint, self._token, data=data)
        return QuickAddResult.from_quick_add_response(task_data)

    def get_project(self, project_id: str) -> Project:
        endpoint = get_rest_url(f"{PROJECTS_ENDPOINT}/{project_id}")
        project = get(self._session, endpoint, self._token)
        return Project.from_dict(project)

    def get_projects(self) -> list[Project]:
        endpoint = get_rest_url(PROJECTS_ENDPOINT)
        projects = get(self._session, endpoint, self._token)
        return [Project.from_dict(obj) for obj in projects]

    def add_project(self, name: str, **kwargs) -> Project:
        endpoint = get_rest_url(PROJECTS_ENDPOINT)
        data: dict[str, Any] = {"name": name}
        data.update(kwargs)
        project = post(self._session, endpoint, self._token, data=data)
        return Project.from_dict(project)

    def update_project(self, project_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{PROJECTS_ENDPOINT}/{project_id}")
        return post(self._session, endpoint, self._token, data=kwargs)

    def delete_project(self, project_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{PROJECTS_ENDPOINT}/{project_id}")
        return delete(self._session, endpoint, self._token, args=kwargs)

    def get_collaborators(self, project_id: str) -> list[Collaborator]:
        endpoint = get_rest_url(
            f"{PROJECTS_ENDPOINT}/{project_id}/{COLLABORATORS_ENDPOINT}"
        )
        collaborators = get(self._session, endpoint, self._token)
        return [Collaborator.from_dict(obj) for obj in collaborators]

    def get_section(self, section_id: str) -> Section:
        endpoint = get_rest_url(f"{SECTIONS_ENDPOINT}/{section_id}")
        section = get(self._session, endpoint, self._token)
        return Section.from_dict(section)

    def get_sections(self, **kwargs) -> list[Section]:
        endpoint = get_rest_url(SECTIONS_ENDPOINT)
        sections = get(self._session, endpoint, self._token, kwargs)
        return [Section.from_dict(obj) for obj in sections]

    def add_section(self, name: str, project_id: str, **kwargs) -> Section:
        endpoint = get_rest_url(SECTIONS_ENDPOINT)
        data = {"name": name, "project_id": project_id}
        data.update(kwargs)
        section = post(self._session, endpoint, self._token, data=data)
        return Section.from_dict(section)

    def update_section(self, section_id: str, name: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{SECTIONS_ENDPOINT}/{section_id}")
        data: dict[str, Any] = {"name": name}
        data.update(kwargs)
        return post(self._session, endpoint, self._token, data=data)

    def delete_section(self, section_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{SECTIONS_ENDPOINT}/{section_id}")
        return delete(self._session, endpoint, self._token, args=kwargs)

    def get_comment(self, comment_id: str) -> Comment:
        endpoint = get_rest_url(f"{COMMENTS_ENDPOINT}/{comment_id}")
        comment = get(self._session, endpoint, self._token)
        return Comment.from_dict(comment)

    def get_comments(self, **kwargs) -> list[Comment]:
        endpoint = get_rest_url(COMMENTS_ENDPOINT)
        comments = get(self._session, endpoint, self._token, kwargs)
        return [Comment.from_dict(obj) for obj in comments]

    def add_comment(self, content: str, **kwargs) -> Comment:
        endpoint = get_rest_url(COMMENTS_ENDPOINT)
        data = {"content": content}
        data.update(kwargs)
        comment = post(self._session, endpoint, self._token, data=data)
        return Comment.from_dict(comment)

    def update_comment(self, comment_id: str, content: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{COMMENTS_ENDPOINT}/{comment_id}")
        data: dict[str, Any] = {"content": content}
        data.update(kwargs)
        return post(self._session, endpoint, self._token, data=data)

    def delete_comment(self, comment_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{COMMENTS_ENDPOINT}/{comment_id}")
        return delete(self._session, endpoint, self._token, args=kwargs)

    def get_label(self, label_id: str) -> Label:
        endpoint = get_rest_url(f"{LABELS_ENDPOINT}/{label_id}")
        label = get(self._session, endpoint, self._token)
        return Label.from_dict(label)

    def get_labels(self) -> list[Label]:
        endpoint = get_rest_url(LABELS_ENDPOINT)
        labels = get(self._session, endpoint, self._token)
        return [Label.from_dict(obj) for obj in labels]

    def add_label(self, name: str, **kwargs) -> Label:
        endpoint = get_rest_url(LABELS_ENDPOINT)
        data = {"name": name}
        data.update(kwargs)
        label = post(self._session, endpoint, self._token, data=data)
        return Label.from_dict(label)

    def update_label(self, label_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{LABELS_ENDPOINT}/{label_id}")
        return post(self._session, endpoint, self._token, data=kwargs)

    def delete_label(self, label_id: str, **kwargs) -> bool:
        endpoint = get_rest_url(f"{LABELS_ENDPOINT}/{label_id}")
        return delete(self._session, endpoint, self._token, args=kwargs)

    def get_shared_labels(self) -> list[str]:
        endpoint = get_rest_url(SHARED_LABELS_ENDPOINT)
        return get(self._session, endpoint, self._token)

    def rename_shared_label(self, name: str, new_name: str) -> bool:
        endpoint = get_rest_url(SHARED_LABELS_RENAME_ENDPOINT)
        data = {"name": name, "new_name": new_name}
        return post(self._session, endpoint, self._token, data=data)

    def remove_shared_label(self, name: str) -> bool:
        endpoint = get_rest_url(SHARED_LABELS_REMOVE_ENDPOINT)
        data = {"name": name}
        return post(self._session, endpoint, self._token, data=data)

    def get_completed_items(
        self,
        project_id: str | None = None,
        section_id: str | None = None,
        item_id: str | None = None,
        last_seen_id: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> CompletedItems:
        endpoint = get_sync_url(COMPLETED_ITEMS_ENDPOINT)
        completed_items = get(
            self._session,
            endpoint,
            self._token,
            {
                "project_id": project_id,
                "section_id": section_id,
                "item_id": item_id,
                "last_seen_id": last_seen_id,
                "limit": limit,
                "cursor": cursor,
            },
        )
        return CompletedItems.from_dict(completed_items)



from __future__ import annotations

from todoist_api_python.headers import create_headers


def test_create_headers_none():
    headers = create_headers()
    assert headers == {}


def test_create_headers_authorization():
    token = "A Token"
    headers = create_headers(token=token)
    assert headers["Authorization"] == f"Bearer {token}"


def test_create_headers_request_id():
    request_id = "12345"
    headers = create_headers(request_id=request_id)
    assert headers["X-Request-Id"] == request_id


def test_create_headers_content_type():
    headers = create_headers(with_content=True)
    assert headers["Content-Type"] == "application/json; charset=utf-8"



from __future__ import annotations

from typing import Any

REST_API_BASE_URL = "https://api.todoist.com/rest/v2"
SYNC_API_BASE_URL = "https://api.todoist.com/sync/v9"
AUTH_BASE_URL = "https://todoist.com"
DEFAULT_TOKEN = "A TOKEN"
DEFAULT_REQUEST_ID = "REQUEST12345"

DEFAULT_DUE_RESPONSE = {
    "date": "2016-09-01",
    "is_recurring": True,
    "datetime": "2016-09-01T09:00:00.00000Z",
    "string": "tomorrow at 12",
    "timezone": "Europe/Moscow",
}

DEFAULT_DURATION_RESPONSE = {
    "amount": 60,
    "unit": "minute",
}

DEFAULT_TASK_RESPONSE = {
    "id": "1234",
    "assigner_id": "2971358",
    "assignee_id": "2423523",
    "project_id": "2203306141",
    "parent_id": "8686843758",
    "section_id": "7025",
    "order": 3,
    "content": "Some Task Content",
    "description": "A description",
    "is_completed": False,
    "is_shared": False,
    "labels": [],
    "priority": 1,
    "comment_count": 0,
    "creator_id": "0",
    "created_at": "2019-01-02T21:00:30.00000Z",
    "url": "https://todoist.com/showTask?id=2995104339",
    "due": DEFAULT_DUE_RESPONSE,
    "duration": DEFAULT_DURATION_RESPONSE,
}

DEFAULT_TASK_RESPONSE_2 = dict(DEFAULT_TASK_RESPONSE)
DEFAULT_TASK_RESPONSE_2["id"] = "5678"

DEFAULT_TASKS_RESPONSE = [
    DEFAULT_TASK_RESPONSE,
    DEFAULT_TASK_RESPONSE_2,
]

DEFAULT_PROJECT_RESPONSE = {
    "id": "1234",
    "name": "Inbox",
    "comment_count": 10,
    "order": 1,
    "color": "red",
    "is_shared": False,
    "parent_id": "5678",
    "is_favorite": False,
    "is_inbox_project": True,
    "is_team_inbox": True,
    "can_assign_tasks": False,
    "url": "https://todoist.com/showProject?id=1234",
    "view_style": "list",
}

DEFAULT_PROJECT_RESPONSE_2 = dict(DEFAULT_PROJECT_RESPONSE)
DEFAULT_PROJECT_RESPONSE_2["id"] = "5678"

DEFAULT_PROJECTS_RESPONSE = [
    DEFAULT_PROJECT_RESPONSE,
    DEFAULT_PROJECT_RESPONSE_2,
]

DEFAULT_COLLABORATOR_RESPONSE = {
    "id": "1234",
    "name": "Alice",
    "email": "alice@example.com",
}

DEFAULT_COLLABORATOR_RESPONSE_2 = dict(DEFAULT_COLLABORATOR_RESPONSE)
DEFAULT_COLLABORATOR_RESPONSE_2["id"] = "5678"

DEFAULT_COLLABORATORS_RESPONSE = [
    DEFAULT_COLLABORATOR_RESPONSE,
    DEFAULT_COLLABORATOR_RESPONSE_2,
]

DEFAULT_SECTION_RESPONSE = {
    "id": "1234",
    "project_id": "4567",
    "name": "A Section",
    "order": 1,
}

DEFAULT_SECTION_RESPONSE_2 = dict(DEFAULT_SECTION_RESPONSE)
DEFAULT_SECTION_RESPONSE_2["id"] = 5678

DEFAULT_SECTIONS_RESPONSE = [
    DEFAULT_SECTION_RESPONSE,
    DEFAULT_SECTION_RESPONSE_2,
]

DEFAULT_ATTACHMENT_RESPONSE = {
    "resource_type": "file",
    "file_name": "File.pdf",
    "file_type": "application/pdf",
    "file_size": 4321,
    "file_url": "https://cdn-domain.tld/path/to/file.pdf",
    "upload_state": "completed",
    "image": "https://cdn-domain.tld/path/to/some_image.jpg",
    "image_width": 1234,
    "image_height": 5678,
    "url": "https://todoist.com",
    "title": "Todoist Website",
}

DEFAULT_COMMENT_RESPONSE: dict[str, Any] = {
    "id": "1234",
    "content": "A comment",
    "posted_at": "2016-09-22T07:00:00.00000Z",
    "task_id": "2345",
    "project_id": "4567",
    "attachment": DEFAULT_ATTACHMENT_RESPONSE,
}

DEFAULT_COMMENT_RESPONSE_2 = dict(DEFAULT_COMMENT_RESPONSE)
DEFAULT_COMMENT_RESPONSE_2["id"] = "5678"
DEFAULT_COMMENT_RESPONSE_2["attachment"] = None

DEFAULT_COMMENTS_RESPONSE = [
    DEFAULT_COMMENT_RESPONSE,
    DEFAULT_COMMENT_RESPONSE_2,
]

DEFAULT_LABEL_RESPONSE = {
    "id": "1234",
    "name": "A label",
    "color": "red",
    "order": 1,
    "is_favorite": True,
}

DEFAULT_LABEL_RESPONSE_2 = dict(DEFAULT_LABEL_RESPONSE)
DEFAULT_LABEL_RESPONSE_2["id"] = "4567"

DEFAULT_LABELS_RESPONSE = [
    DEFAULT_LABEL_RESPONSE,
    DEFAULT_LABEL_RESPONSE_2,
]

DEFAULT_AUTH_RESPONSE = {
    "access_token": "1234",
    "state": "somestate",
}

DEFAULT_ITEM_RESPONSE = {
    "id": "2995104339",
    "user_id": "2671355",
    "project_id": "2203306141",
    "content": "Buy Milk",
    "description": "",
    "priority": 1,
    "due": DEFAULT_DUE_RESPONSE,
    "child_order": 1,
    "day_order": -1,
    "collapsed": False,
    "labels": ["Food", "Shopping"],
    "added_by_uid": "2671355",
    "assigned_by_uid": "2671355",
    "checked": False,
    "is_deleted": False,
    "added_at": "2014-09-26T08:25:05.000000Z",
}

DEFAULT_ITEM_COMPLETED_INFO_RESPONSE = {"item_id": "2995104339", "completed_items": 12}

DEFAULT_COMPLETED_ITEMS_RESPONSE = {
    "items": [DEFAULT_ITEM_RESPONSE],
    "completed_info": [DEFAULT_ITEM_COMPLETED_INFO_RESPONSE],
    "total": 22,
    "next_cursor": "k85gVI5ZAs8AAAABFoOzAQ",
    "has_more": True,
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs, urlparse

import pytest
import responses

from tests.data.test_defaults import SYNC_API_BASE_URL
from tests.utils.test_utils import assert_auth_header
from todoist_api_python.endpoints import COMPLETED_ITEMS_ENDPOINT

if TYPE_CHECKING:
    from todoist_api_python.api import TodoistAPI
    from todoist_api_python.api_async import TodoistAPIAsync
    from todoist_api_python.models import CompletedItems


@pytest.mark.asyncio
async def test_get_completed_items(
    todoist_api: TodoistAPI,
    todoist_api_async: TodoistAPIAsync,
    requests_mock: responses.RequestsMock,
    default_completed_items_response: dict[str, Any],
    default_completed_items: CompletedItems,
) -> None:
    project_id = "1234"
    section_id = "5678"
    item_id = "90ab"
    last_seen_id = "cdef"
    limit = 30
    cursor = "ghij"

    def assert_query(url):
        queries = parse_qs(urlparse(url).query)
        assert queries.get("project_id") == [project_id]
        assert queries.get("section_id") == [section_id]
        assert queries.get("item_id") == [item_id]
        assert queries.get("last_seen_id") == [last_seen_id]
        assert queries.get("limit") == [str(limit)]
        assert queries.get("cursor") == [cursor]

    expected_endpoint = f"{SYNC_API_BASE_URL}/{COMPLETED_ITEMS_ENDPOINT}"

    requests_mock.add(
        responses.GET,
        expected_endpoint,
        json=default_completed_items_response,
        status=200,
    )

    completed_items = todoist_api.get_completed_items(
        project_id, section_id, item_id, last_seen_id, limit, cursor
    )

    assert len(requests_mock.calls) == 1
    assert_auth_header(requests_mock.calls[0].request)
    assert_query(requests_mock.calls[0].request.url)
    assert completed_items == default_completed_items

    completed_items = await todoist_api_async.get_completed_items(
        project_id, section_id, item_id, last_seen_id, limit, cursor
    )

    assert len(requests_mock.calls) == 2
    assert_auth_header(requests_mock.calls[1].request)
    assert_query(requests_mock.calls[1].request.url)
    assert completed_items == default_completed_items



from __future__ import annotations

from unittest.mock import MagicMock, patch

import requests

from tests.data.test_defaults import DEFAULT_TOKEN
from tests.utils.test_utils import get_todoist_api_patch
from todoist_api_python.api import TodoistAPI
from todoist_api_python.api_async import TodoistAPIAsync


@patch(get_todoist_api_patch(TodoistAPI.__init__))
def test_constructs_api_with_token(sync_api_constructor: MagicMock):
    sync_api_constructor.return_value = None
    TodoistAPIAsync(DEFAULT_TOKEN)

    sync_api_constructor.assert_called_once_with(DEFAULT_TOKEN, None)


@patch(get_todoist_api_patch(TodoistAPI.__init__))
def test_constructs_api_with_token_and_session(sync_api_constructor: MagicMock):
    sync_api_constructor.return_value = None
    session = requests.Session()
    TodoistAPIAsync(DEFAULT_TOKEN, session)
    sync_api_constructor.assert_called_once_with(DEFAULT_TOKEN, session)



from __future__ import annotations

from urllib.parse import urljoin

BASE_URL = "https://api.todoist.com"
AUTH_BASE_URL = "https://todoist.com"
SYNC_VERSION = "v9"
REST_VERSION = "v2"

SYNC_API = urljoin(BASE_URL, f"/sync/{SYNC_VERSION}/")
REST_API = urljoin(BASE_URL, f"/rest/{REST_VERSION}/")


TASKS_ENDPOINT = "tasks"
PROJECTS_ENDPOINT = "projects"
COLLABORATORS_ENDPOINT = "collaborators"
SECTIONS_ENDPOINT = "sections"
COMMENTS_ENDPOINT = "comments"
LABELS_ENDPOINT = "labels"
SHARED_LABELS_ENDPOINT = "labels/shared"
SHARED_LABELS_RENAME_ENDPOINT = f"{SHARED_LABELS_ENDPOINT}/rename"
SHARED_LABELS_REMOVE_ENDPOINT = f"{SHARED_LABELS_ENDPOINT}/remove"
QUICK_ADD_ENDPOINT = "quick/add"

AUTHORIZE_ENDPOINT = "oauth/authorize"
TOKEN_ENDPOINT = "oauth/access_token"
REVOKE_TOKEN_ENDPOINT = "access_tokens/revoke"

COMPLETED_ITEMS_ENDPOINT = "archive/items"


def get_rest_url(relative_path: str) -> str:
    return urljoin(REST_API, relative_path)


def get_sync_url(relative_path: str) -> str:
    return urljoin(SYNC_API, relative_path)


def get_auth_url(relative_path: str) -> str:
    return urljoin(AUTH_BASE_URL, relative_path)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

import pytest
import responses

from tests.data.test_defaults import AUTH_BASE_URL
from todoist_api_python.authentication import (
    get_auth_token,
    get_auth_token_async,
    get_authentication_url,
    revoke_auth_token,
    revoke_auth_token_async,
)
from todoist_api_python.endpoints import SYNC_API

if TYPE_CHECKING:
    from todoist_api_python.models import AuthResult


def test_get_authentication_url():
    client_id = "123"
    scopes = ["task:add", "data:read", "project:delete"]
    state = "456"
    params = (
        f"client_id={client_id}&scope={scopes[0]},{scopes[1]},{scopes[2]}&state={state}"
    )
    query = quote(params, safe="=&")
    expected_url = f"{AUTH_BASE_URL}/oauth/authorize?{query}"

    url = get_authentication_url(client_id, scopes, state)

    assert url == expected_url


@pytest.mark.asyncio
async def test_get_auth_token(
    requests_mock: responses.RequestsMock,
    default_auth_response: dict[str, Any],
    default_auth_result: AuthResult,
):
    client_id = "123"
    client_secret = "456"
    code = "789"

    expected_payload = json.dumps(
        {"client_id": client_id, "client_secret": client_secret, "code": code}
    )

    requests_mock.add(
        responses.POST,
        f"{AUTH_BASE_URL}/oauth/access_token",
        json=default_auth_response,
        status=200,
    )

    auth_result = get_auth_token(client_id, client_secret, code)

    assert len(requests_mock.calls) == 1
    assert requests_mock.calls[0].request.body == expected_payload
    assert auth_result == default_auth_result

    auth_result = await get_auth_token_async(client_id, client_secret, code)

    assert len(requests_mock.calls) == 2
    assert requests_mock.calls[1].request.body == expected_payload
    assert auth_result == default_auth_result


@pytest.mark.asyncio
async def test_revoke_auth_token(
    requests_mock: responses.RequestsMock,
):
    client_id = "123"
    client_secret = "456"
    token = "AToken"

    expected_payload = json.dumps(
        {"client_id": client_id, "client_secret": client_secret, "access_token": token}
    )

    requests_mock.add(
        responses.POST,
        f"{SYNC_API}access_tokens/revoke",
        status=204,
    )

    result = revoke_auth_token(client_id, client_secret, token)

    assert len(requests_mock.calls) == 1
    assert requests_mock.calls[0].request.body == expected_payload
    assert result is True

    result = await revoke_auth_token_async(client_id, client_secret, token)

    assert len(requests_mock.calls) == 2
    assert requests_mock.calls[1].request.body == expected_payload
    assert result is True



from __future__ import annotations

from urllib.parse import urlencode

import requests
from requests import Session

from todoist_api_python.endpoints import (
    AUTHORIZE_ENDPOINT,
    REVOKE_TOKEN_ENDPOINT,
    TOKEN_ENDPOINT,
    get_auth_url,
    get_sync_url,
)
from todoist_api_python.http_requests import post
from todoist_api_python.models import AuthResult
from todoist_api_python.utils import run_async


def get_auth_token(
    client_id: str, client_secret: str, code: str, session: Session | None = None
) -> AuthResult:
    endpoint = get_auth_url(TOKEN_ENDPOINT)
    session = session or requests.Session()
    payload = {"client_id": client_id, "client_secret": client_secret, "code": code}
    response = post(session=session, url=endpoint, data=payload)

    return AuthResult.from_dict(response)


async def get_auth_token_async(
    client_id: str, client_secret: str, code: str
) -> AuthResult:
    return await run_async(lambda: get_auth_token(client_id, client_secret, code))


def revoke_auth_token(
    client_id: str, client_secret: str, token: str, session: Session | None = None
) -> bool:
    endpoint = get_sync_url(REVOKE_TOKEN_ENDPOINT)
    session = session or requests.Session()
    payload = {
        "client_id": client_id,
        "client_secret": client_secret,
        "access_token": token,
    }
    response = post(session=session, url=endpoint, data=payload)

    return response


async def revoke_auth_token_async(
    client_id: str, client_secret: str, token: str
) -> bool:
    return await run_async(lambda: revoke_auth_token(client_id, client_secret, token))


class ArgumentError(Exception):
    pass


def get_authentication_url(client_id: str, scopes: list[str], state: str) -> str:
    if len(scopes) == 0:
        raise ArgumentError("At least one authorization scope should be requested.")

    query = {"client_id": client_id, "scope": ",".join(scopes), "state": state}

    auth_url = get_auth_url(AUTHORIZE_ENDPOINT)

    return f"{auth_url}?{urlencode(query)}"



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from todoist_api_python.headers import create_headers

if TYPE_CHECKING:
    from requests import Session


def get(
    session: Session,
    url: str,
    token: str | None = None,
    params: dict[str, Any] | None = None,
):
    response = session.get(url, params=params, headers=create_headers(token=token))

    if response.status_code == 200:
        return response.json()

    response.raise_for_status()
    return response.ok


def post(
    session: Session,
    url: str,
    token: str | None = None,
    data: dict[str, Any] | None = None,
):
    request_id = data.pop("request_id", None) if data else None

    headers = create_headers(
        token=token, with_content=bool(data), request_id=request_id
    )

    response = session.post(
        url,
        headers=headers,
        data=json.dumps(data) if data else None,
    )

    if response.status_code == 200:
        return response.json()

    response.raise_for_status()
    return response.ok


def delete(
    session: Session,
    url: str,
    token: str | None = None,
    args: dict[str, Any] | None = None,
):
    request_id = args.pop("request_id", None) if args else None

    headers = create_headers(token=token, request_id=request_id)

    response = session.delete(
        url,
        headers=headers,
    )

    response.raise_for_status()
    return response.ok



from __future__ import annotations

from tests.data.quick_add_responses import (
    QUICK_ADD_RESPONSE_FULL,
    QUICK_ADD_RESPONSE_MINIMAL,
)
from tests.data.test_defaults import (
    DEFAULT_ATTACHMENT_RESPONSE,
    DEFAULT_COLLABORATOR_RESPONSE,
    DEFAULT_COMMENT_RESPONSE,
    DEFAULT_COMPLETED_ITEMS_RESPONSE,
    DEFAULT_DUE_RESPONSE,
    DEFAULT_DURATION_RESPONSE,
    DEFAULT_ITEM_COMPLETED_INFO_RESPONSE,
    DEFAULT_ITEM_RESPONSE,
    DEFAULT_LABEL_RESPONSE,
    DEFAULT_PROJECT_RESPONSE,
    DEFAULT_SECTION_RESPONSE,
    DEFAULT_TASK_RESPONSE,
)
from todoist_api_python.models import (
    Attachment,
    AuthResult,
    Collaborator,
    Comment,
    CompletedItems,
    Due,
    Duration,
    Item,
    ItemCompletedInfo,
    Label,
    Project,
    QuickAddResult,
    Section,
    Task,
)

unexpected_data = {"unexpected_key": "some value"}


def test_project_from_dict():
    sample_data = dict(DEFAULT_PROJECT_RESPONSE)
    sample_data.update(unexpected_data)

    project = Project.from_dict(sample_data)

    assert project.id == sample_data["id"]
    assert project.color == sample_data["color"]
    assert project.comment_count == sample_data["comment_count"]
    assert project.is_favorite == sample_data["is_favorite"]
    assert project.name == sample_data["name"]
    assert project.is_shared == sample_data["is_shared"]
    assert project.url == sample_data["url"]
    assert project.is_inbox_project == sample_data["is_inbox_project"]
    assert project.is_team_inbox == sample_data["is_team_inbox"]
    assert project.order == sample_data["order"]
    assert project.parent_id == sample_data["parent_id"]
    assert project.view_style == sample_data["view_style"]
    assert project.can_assign_tasks == sample_data["can_assign_tasks"]


def test_section_from_dict():
    sample_data = dict(DEFAULT_SECTION_RESPONSE)
    sample_data.update(unexpected_data)

    section = Section.from_dict(sample_data)

    assert section.id == sample_data["id"]
    assert section.name == sample_data["name"]
    assert section.order == sample_data["order"]
    assert section.project_id == sample_data["project_id"]


def test_due_from_dict():
    sample_data = dict(DEFAULT_DUE_RESPONSE)
    sample_data.update(unexpected_data)

    due = Due.from_dict(sample_data)

    assert due.date == sample_data["date"]
    assert due.is_recurring == sample_data["is_recurring"]
    assert due.string == sample_data["string"]
    assert due.datetime == sample_data["datetime"]
    assert due.timezone == sample_data["timezone"]


def test_duration_from_dict():
    sample_data = dict(DEFAULT_DURATION_RESPONSE)
    sample_data.update(unexpected_data)

    duration = Duration.from_dict(sample_data)

    assert duration.amount == sample_data["amount"]
    assert duration.unit == sample_data["unit"]


def test_task_from_dict():
    sample_data = dict(DEFAULT_TASK_RESPONSE)
    sample_data.update(unexpected_data)

    task = Task.from_dict(sample_data)

    assert task.comment_count == sample_data["comment_count"]
    assert task.is_completed == sample_data["is_completed"]
    assert task.content == sample_data["content"]
    assert task.created_at == sample_data["created_at"]
    assert task.creator_id == sample_data["creator_id"]
    assert task.id == sample_data["id"]
    assert task.project_id == sample_data["project_id"]
    assert task.section_id == sample_data["section_id"]
    assert task.priority == sample_data["priority"]
    assert task.url == sample_data["url"]
    assert task.assignee_id == sample_data["assignee_id"]
    assert task.assigner_id == sample_data["assigner_id"]
    assert task.due == Due.from_dict(sample_data["due"])
    assert task.labels == sample_data["labels"]
    assert task.order == sample_data["order"]
    assert task.parent_id == sample_data["parent_id"]
    assert task.duration == Duration.from_dict(sample_data["duration"])


def test_task_to_dict():
    sample_data = dict(DEFAULT_TASK_RESPONSE)
    sample_data.update(unexpected_data)

    task = Task.from_dict(sample_data).to_dict()

    assert task["comment_count"] == sample_data["comment_count"]
    assert task["is_completed"] == sample_data["is_completed"]
    assert task["content"] == sample_data["content"]
    assert task["created_at"] == sample_data["created_at"]
    assert task["creator_id"] == sample_data["creator_id"]
    assert task["id"] == sample_data["id"]
    assert task["project_id"] == sample_data["project_id"]
    assert task["section_id"] == sample_data["section_id"]
    assert task["priority"] == sample_data["priority"]
    assert task["url"] == sample_data["url"]
    assert task["assignee_id"] == sample_data["assignee_id"]
    assert task["assigner_id"] == sample_data["assigner_id"]
    for key in task["due"]:
        assert task["due"][key] == sample_data["due"][key]
    assert task["labels"] == sample_data["labels"]
    assert task["order"] == sample_data["order"]
    assert task["parent_id"] == sample_data["parent_id"]
    for key in task["duration"]:
        assert task["duration"][key] == sample_data["duration"][key]


def test_collaborator_from_dict():
    sample_data = dict(DEFAULT_COLLABORATOR_RESPONSE)
    sample_data.update(unexpected_data)

    collaborator = Collaborator.from_dict(sample_data)

    assert collaborator.id == sample_data["id"]
    assert collaborator.email == sample_data["email"]
    assert collaborator.name == sample_data["name"]


def test_attachment_from_dict():
    sample_data = dict(DEFAULT_ATTACHMENT_RESPONSE)
    sample_data.update(unexpected_data)

    attachment = Attachment.from_dict(sample_data)

    assert attachment.resource_type == sample_data["resource_type"]
    assert attachment.file_name == sample_data["file_name"]
    assert attachment.file_size == sample_data["file_size"]
    assert attachment.file_type == sample_data["file_type"]
    assert attachment.file_url == sample_data["file_url"]
    assert attachment.upload_state == sample_data["upload_state"]
    assert attachment.image == sample_data["image"]
    assert attachment.image_width == sample_data["image_width"]
    assert attachment.image_height == sample_data["image_height"]
    assert attachment.url == sample_data["url"]
    assert attachment.title == sample_data["title"]


def test_comment_from_dict():
    sample_data = dict(DEFAULT_COMMENT_RESPONSE)
    sample_data.update(unexpected_data)

    comment = Comment.from_dict(sample_data)

    assert comment.id == sample_data["id"]
    assert comment.content == sample_data["content"]
    assert comment.posted_at == sample_data["posted_at"]
    assert comment.task_id == sample_data["task_id"]
    assert comment.project_id == sample_data["project_id"]
    assert comment.attachment == Attachment.from_dict(sample_data["attachment"])


def test_label_from_dict():
    sample_data = dict(DEFAULT_LABEL_RESPONSE)
    sample_data.update(unexpected_data)

    label = Label.from_dict(sample_data)

    assert label.id == sample_data["id"]
    assert label.name == sample_data["name"]
    assert label.color == sample_data["color"]
    assert label.order == sample_data["order"]
    assert label.is_favorite == sample_data["is_favorite"]


def test_quick_add_result_minimal():
    sample_data = dict(QUICK_ADD_RESPONSE_MINIMAL)
    sample_data.update(unexpected_data)

    quick_add_result = QuickAddResult.from_quick_add_response(sample_data)

    assert quick_add_result.task.comment_count == 0
    assert quick_add_result.task.is_completed is False
    assert quick_add_result.task.content == "some task"
    assert quick_add_result.task.created_at == "2021-02-05T11:02:56.00000Z"
    assert quick_add_result.task.creator_id == "21180723"
    assert quick_add_result.task.id == "4554989047"
    assert quick_add_result.task.project_id == "2203108698"
    assert quick_add_result.task.section_id is None
    assert quick_add_result.task.priority == 1
    assert quick_add_result.task.url == "https://todoist.com/showTask?id=4554989047"
    assert quick_add_result.task.assignee_id is None
    assert quick_add_result.task.assigner_id is None
    assert quick_add_result.task.due is None
    assert quick_add_result.task.labels == []
    assert quick_add_result.task.order == 6
    assert quick_add_result.task.parent_id is None
    assert quick_add_result.task.sync_id is None

    assert quick_add_result.resolved_assignee_name is None
    assert quick_add_result.resolved_label_names == []
    assert quick_add_result.resolved_project_name is None
    assert quick_add_result.resolved_section_name is None


def test_quick_add_result_full():
    sample_data = dict(QUICK_ADD_RESPONSE_FULL)
    sample_data.update(unexpected_data)

    quick_add_result = QuickAddResult.from_quick_add_response(sample_data)

    assert quick_add_result.task.comment_count == 0
    assert quick_add_result.task.is_completed is False
    assert quick_add_result.task.content == "some task"
    assert quick_add_result.task.created_at == "2021-02-05T11:04:54.00000Z"
    assert quick_add_result.task.creator_id == "21180723"
    assert quick_add_result.task.id == "4554993687"
    assert quick_add_result.task.project_id == "2257514220"
    assert quick_add_result.task.section_id == "2232454220"
    assert quick_add_result.task.priority == 1
    assert (
        quick_add_result.task.url
        == "https://todoist.com/showTask?id=4554993687&sync_id=4554993687"
    )
    assert quick_add_result.task.assignee_id == "29172386"
    assert quick_add_result.task.assigner_id == "21180723"
    assert quick_add_result.task.due.date == "2021-02-06T11:00:00.00000Z"
    assert quick_add_result.task.due.is_recurring is False
    assert quick_add_result.task.due.string == "Feb 6 11:00 AM"
    assert quick_add_result.task.due.datetime == "2021-02-06T11:00:00.00000Z"
    assert quick_add_result.task.due.timezone == "Europe/London"
    assert quick_add_result.task.labels == ["Label1", "Label2"]
    assert quick_add_result.task.order == 1
    assert quick_add_result.task.parent_id is None
    assert quick_add_result.task.sync_id == "4554993687"

    assert quick_add_result.resolved_assignee_name == "Some Guy"
    assert quick_add_result.resolved_label_names == ["Label1", "Label2"]
    assert quick_add_result.resolved_project_name == "test"
    assert quick_add_result.resolved_section_name == "A section"


def test_quick_add_broken_data():
    none_attribute = QUICK_ADD_RESPONSE_FULL.copy()
    missing_attribute = QUICK_ADD_RESPONSE_FULL.copy()

    none_attribute["meta"]["project"] = None
    none_attribute["meta"]["assignee"] = None
    none_attribute["meta"]["section"] = None

    del missing_attribute["meta"]["project"]
    del missing_attribute["meta"]["assignee"]
    del missing_attribute["meta"]["section"]

    for quick_add_responses in [none_attribute, missing_attribute]:
        sample_data = dict(quick_add_responses)
        sample_data.update(unexpected_data)

        quick_add_result = QuickAddResult.from_quick_add_response(sample_data)

        assert quick_add_result.task.comment_count == 0
        assert quick_add_result.task.is_completed is False
        assert quick_add_result.task.content == "some task"
        assert quick_add_result.task.created_at == "2021-02-05T11:04:54.00000Z"
        assert quick_add_result.task.creator_id == "21180723"
        assert quick_add_result.task.id == "4554993687"
        assert quick_add_result.task.project_id == "2257514220"
        assert quick_add_result.task.section_id == "2232454220"
        assert quick_add_result.task.priority == 1
        assert (
            quick_add_result.task.url
            == "https://todoist.com/showTask?id=4554993687&sync_id=4554993687"
        )
        assert quick_add_result.task.assignee_id == "29172386"
        assert quick_add_result.task.assigner_id == "21180723"
        assert quick_add_result.task.due.date == "2021-02-06T11:00:00.00000Z"
        assert quick_add_result.task.due.is_recurring is False
        assert quick_add_result.task.due.string == "Feb 6 11:00 AM"
        assert quick_add_result.task.due.datetime == "2021-02-06T11:00:00.00000Z"
        assert quick_add_result.task.due.timezone == "Europe/London"
        assert quick_add_result.task.labels == ["Label1", "Label2"]
        assert quick_add_result.task.order == 1
        assert quick_add_result.task.parent_id is None
        assert quick_add_result.task.sync_id == "4554993687"

        assert quick_add_result.resolved_assignee_name is None
        assert quick_add_result.resolved_label_names == ["Label1", "Label2"]
        assert quick_add_result.resolved_project_name is None
        assert quick_add_result.resolved_section_name is None


def test_auth_result_from_dict():
    token = "123"
    state = "456"
    sample_data = {"access_token": token, "state": state}
    sample_data.update(unexpected_data)

    auth_result = AuthResult.from_dict(sample_data)

    assert auth_result.access_token == token
    assert auth_result.state == state


def test_item_from_dict():
    sample_data = dict(DEFAULT_ITEM_RESPONSE)
    sample_data.update(unexpected_data)

    item = Item.from_dict(sample_data)

    assert item.id == "2995104339"
    assert item.user_id == "2671355"
    assert item.project_id == "2203306141"
    assert item.content == "Buy Milk"
    assert item.description == ""
    assert item.priority == 1
    assert item.due.date == DEFAULT_DUE_RESPONSE["date"]
    assert item.due.is_recurring == DEFAULT_DUE_RESPONSE["is_recurring"]
    assert item.due.string == DEFAULT_DUE_RESPONSE["string"]
    assert item.due.datetime == DEFAULT_DUE_RESPONSE["datetime"]
    assert item.due.timezone == DEFAULT_DUE_RESPONSE["timezone"]
    assert item.parent_id is None
    assert item.child_order == 1
    assert item.section_id is None
    assert item.day_order == -1
    assert item.collapsed is False
    assert item.labels == ["Food", "Shopping"]
    assert item.added_by_uid == "2671355"
    assert item.assigned_by_uid == "2671355"
    assert item.responsible_uid is None
    assert item.checked is False
    assert item.is_deleted is False
    assert item.sync_id is None
    assert item.added_at == "2014-09-26T08:25:05.000000Z"


def test_item_completed_info_from_dict():
    sample_data = dict(DEFAULT_ITEM_COMPLETED_INFO_RESPONSE)
    sample_data.update(unexpected_data)

    info = ItemCompletedInfo.from_dict(sample_data)

    assert info.item_id == "2995104339"
    assert info.completed_items == 12


def test_completed_items_from_dict():
    sample_data = dict(DEFAULT_COMPLETED_ITEMS_RESPONSE)
    sample_data.update(unexpected_data)

    completed_items = CompletedItems.from_dict(sample_data)

    assert completed_items.total == 22
    assert completed_items.next_cursor == "k85gVI5ZAs8AAAABFoOzAQ"
    assert completed_items.has_more is True
    assert len(completed_items.items) == 1
    assert completed_items.items[0].id == "2995104339"
    assert completed_items.items[0].user_id == "2671355"
    assert completed_items.items[0].project_id == "2203306141"
    assert completed_items.items[0].content == "Buy Milk"
    assert completed_items.items[0].description == ""
    assert completed_items.items[0].priority == 1
    assert completed_items.items[0].due.date == DEFAULT_DUE_RESPONSE["date"]
    assert (
        completed_items.items[0].due.is_recurring
        == DEFAULT_DUE_RESPONSE["is_recurring"]
    )
    assert completed_items.items[0].due.string == DEFAULT_DUE_RESPONSE["string"]
    assert completed_items.items[0].due.datetime == DEFAULT_DUE_RESPONSE["datetime"]
    assert completed_items.items[0].due.timezone == DEFAULT_DUE_RESPONSE["timezone"]
    assert completed_items.items[0].parent_id is None
    assert completed_items.items[0].child_order == 1
    assert completed_items.items[0].section_id is None
    assert completed_items.items[0].day_order == -1
    assert completed_items.items[0].collapsed is False
    assert completed_items.items[0].labels == ["Food", "Shopping"]
    assert completed_items.items[0].added_by_uid == "2671355"
    assert completed_items.items[0].assigned_by_uid == "2671355"
    assert completed_items.items[0].responsible_uid is None
    assert completed_items.items[0].checked is False
    assert completed_items.items[0].is_deleted is False
    assert completed_items.items[0].sync_id is None
    assert completed_items.items[0].added_at == "2014-09-26T08:25:05.000000Z"
    assert len(completed_items.completed_info) == 1
    assert completed_items.completed_info[0].item_id == "2995104339"
    assert completed_items.completed_info[0].completed_items == 12
